import unittest
import isotp
from . import unittest_logging
from . import tools
import time
import math
import socket

@unittest.skipIf(tools.check_isotp_socket_possible() == False, 'Cannot test IsoTP socket. %s' % tools.isotp_socket_impossible_reason())
class TestSocket(unittest.TestCase):

    def setUp(self):
        self.socket_list = []

    def tearDown(self):
        for socket in self.socket_list:
            socket.close()

    def make_socket(self, *args, **kwargs):
        s = isotp.socket(*args, **kwargs)
        self.socket_list.append(s)
        return s

    def test_read_write_opts(self):
        s = self.make_socket()
        s.get_fc_opts()
        s.get_ll_opts()
        s.get_opts()

        s.set_fc_opts(stmin=11, bs=22, wftmax=33)
        o = s.get_fc_opts()
        self.assertEqual(o.stmin, 11)
        self.assertEqual(o.bs, 22)
        self.assertEqual(o.wftmax, 33)

        s.set_opts(optflag=0xFFFF, frame_txtime=0x55, ext_address=0x11, txpad=0x22, rxpad=0x33, rx_ext_address=0x44)
        o = s.get_opts()

        self.assertEqual(o.optflag, 0xFFFF)
        self.assertEqual(o.frame_txtime, 0x55)
        self.assertEqual(o.ext_address, 0x11)
        self.assertEqual(o.txpad, 0x22)
        self.assertEqual(o.rxpad, 0x33)
        self.assertEqual(o.rx_ext_address, 0x44)

    def test_receive_transmit(self):
        payload = b'a'*200
        (txid, rxid) = tools.get_next_can_id_pair()
        s1 = self.make_socket()
        s2 = self.make_socket(timeout=0.2)
        addr1 = isotp.Address(isotp.AddressingMode.Normal_11bits, txid=txid, rxid=rxid)
        addr2 = isotp.Address(isotp.AddressingMode.Normal_11bits, txid=rxid, rxid=txid)
        s1.bind(interface = tools.get_test_interface(), address=addr1)
        s2.bind(interface = tools.get_test_interface(), address=addr2)
        s1.send(payload)
        t1 = time.time()
        payload2 = s2.recv()
        self.assertEqual(payload, payload2)

    def test_receive_transmit_fc_opts(self):
        (txid, rxid) = tools.get_next_can_id_pair()
        payload = b'a'*200
        ncf = math.ceil(max(len(payload)-6,0)/7)
        expected_time = ncf * 0.1
        s1 = self.make_socket()
        s2 = self.make_socket(timeout=2*expected_time)
        addr1 = isotp.Address(isotp.AddressingMode.Normal_11bits, txid=txid, rxid=rxid)
        addr2 = isotp.Address(isotp.AddressingMode.Normal_11bits, txid=rxid, rxid=txid)

        s2.set_fc_opts(stmin=100, bs=5)

        s1.bind(interface = tools.get_test_interface(), address=addr1)
        s2.bind(interface = tools.get_test_interface(), address=addr2)
        s1.send(payload)
        t1 = time.time()
        payload2 = s2.recv()
        diff = time.time() - t1
        self.assertEqual(payload, payload2)
        self.assertGreater(diff, expected_time*0.9)
        self.assertLess(diff, expected_time*1.1)
        
    def test_addressing_normal_11bits(self):
        addr = isotp.Address(isotp.AddressingMode.Normal_11bits, txid=0x123, rxid=0x456)
        s = self.make_socket()
        s.bind(tools.get_test_interface(), addr)
        (interface, rxid, txid) = s._socket.getsockname()
        opts = s.get_opts()
        self.assertEqual(txid, 0x123)
        self.assertEqual(rxid, 0x456)
        self.assertFalse(opts.rx_ext_address)
        self.assertFalse(opts.ext_address)

    def test_addressing_normal_29bits(self):
        addr = isotp.Address(isotp.AddressingMode.Normal_29bits, txid=0x123456, rxid=0x789ABC)
        s = self.make_socket()
        s.bind(tools.get_test_interface(), addr)
        (interface, rxid, txid) = s._socket.getsockname()
        opts = s.get_opts()
        self.assertEqual(txid, 0x123456 | socket.CAN_EFF_FLAG)
        self.assertEqual(rxid, 0x789ABC | socket.CAN_EFF_FLAG)
        self.assertEqual(opts.optflag & isotp.socket.flags.EXTEND_ADDR, 0)
        self.assertEqual(opts.optflag & isotp.socket.flags.RX_EXT_ADDR, 0)
        

    def test_addressing_normal_fixed_29bits(self):
        addr = isotp.Address(isotp.AddressingMode.NormalFixed_29bits, source_address=0xaa, target_address=0x55)
        s = self.make_socket()
        s.bind(tools.get_test_interface(), addr)
        (interface, rxid, txid) = s._socket.getsockname()
        opts = s.get_opts()
        self.assertEqual(txid, 0x18DA55AA | socket.CAN_EFF_FLAG)
        self.assertEqual(rxid, 0x18DAAA55 | socket.CAN_EFF_FLAG)
        self.assertEqual(opts.optflag & isotp.socket.flags.EXTEND_ADDR, 0)
        self.assertEqual(opts.optflag & isotp.socket.flags.RX_EXT_ADDR, 0)

    def test_addressing_extended_11bits(self):
        addr = isotp.Address(isotp.AddressingMode.Extended_11bits, txid=0x123, rxid=0x456, source_address=0xAA, target_address=0x55)
        s = self.make_socket()
        s.bind(tools.get_test_interface(), addr)
        (interface, rxid, txid) = s._socket.getsockname()
        opts = s.get_opts()
        self.assertEqual(txid, 0x123)
        self.assertEqual(rxid, 0x456)
        self.assertNotEqual(opts.optflag & isotp.socket.flags.EXTEND_ADDR, 0)
        self.assertNotEqual(opts.optflag & isotp.socket.flags.RX_EXT_ADDR, 0)
        self.assertEqual(opts.ext_address , 0x55)
        self.assertEqual(opts.rx_ext_address, 0xAA)

    def test_addressing_extended_29bits(self):
        addr = isotp.Address(isotp.AddressingMode.Extended_29bits, txid=0x123456, rxid=0x789abc, source_address=0xAA, target_address=0x55)
        s = self.make_socket()
        s.bind(tools.get_test_interface(), addr)
        (interface, rxid, txid) = s._socket.getsockname()
        opts = s.get_opts()
        self.assertEqual(txid, 0x123456 | socket.CAN_EFF_FLAG)
        self.assertEqual(rxid, 0x789abc | socket.CAN_EFF_FLAG)
        self.assertNotEqual(opts.optflag & isotp.socket.flags.EXTEND_ADDR, 0)
        self.assertNotEqual(opts.optflag & isotp.socket.flags.RX_EXT_ADDR, 0)
        self.assertEqual(opts.ext_address , 0x55)
        self.assertEqual(opts.rx_ext_address, 0xAA)

    def test_addressing_mixed_11bits(self):
        addr = isotp.Address(isotp.AddressingMode.Mixed_11bits, txid=0x123, rxid=0x456, address_extension=0x99)
        s = self.make_socket()
        s.bind(tools.get_test_interface(), addr)
        (interface, rxid, txid) = s._socket.getsockname()
        opts = s.get_opts()
        self.assertEqual(txid, 0x123)
        self.assertEqual(rxid, 0x456)
        self.assertNotEqual(opts.optflag & isotp.socket.flags.EXTEND_ADDR, 0)
        self.assertNotEqual(opts.optflag & isotp.socket.flags.RX_EXT_ADDR, 0)
        self.assertEqual(opts.ext_address , 0x99)
        self.assertEqual(opts.rx_ext_address, 0x99)

    def test_addressing_mixed_29bits(self):
        addr = isotp.Address(isotp.AddressingMode.Mixed_29bits, source_address=0xAA, target_address=0x55, address_extension=0x99)
        s = self.make_socket()
        s.bind(tools.get_test_interface(), addr)
        (interface, rxid, txid) = s._socket.getsockname()
        opts = s.get_opts()
        self.assertEqual(txid, 0x18CE55AA | socket.CAN_EFF_FLAG)
        self.assertEqual(rxid, 0x18CEAA55 | socket.CAN_EFF_FLAG)
        self.assertNotEqual(opts.optflag & isotp.socket.flags.EXTEND_ADDR, 0)
        self.assertNotEqual(opts.optflag & isotp.socket.flags.RX_EXT_ADDR, 0)
        self.assertEqual(opts.ext_address , 0x99)
        self.assertEqual(opts.rx_ext_address, 0x99)




class IsoTpError(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)

class FlowControlTimeoutError(IsoTpError):
    """
    Happens when the senders fails to sends a Flow Control message in time. 
    Refer to TransportLayer parameter :ref:`rx_flowcontrol_timeout<param_rx_flowcontrol_timeout>`
    """
    pass
class ConsecutiveFrameTimeoutError(IsoTpError):
    """
    Happens when the senders fails to sends a Consecutive Frame message in time. 
    Refer to TransportLayer parameter :ref:`rx_consecutive_frame_timeout<param_rx_consecutive_frame_timeout>`
    """
    pass
class InvalidCanDataError(IsoTpError):
    """
    Happens when a CAN message that cannot be decoded as valid First Frame, Consecutive Frame, Single Frame or Flow Control PDU is received.
    """
    pass
class UnexpectedFlowControlError(IsoTpError):
    """
    Happens when a Flow Control message is received and was not expected
    """
    pass
class UnexpectedConsecutiveFrameError(IsoTpError):
    """
    Happens when a Consecutive Frame message is received and was not expected
    """
    pass
class ReceptionInterruptedWithSingleFrameError(IsoTpError):
    """
    Happens when the reception of a multi packet message reception is interrupted with a new Single Frame PDU.
    """
    pass
class ReceptionInterruptedWithFirstFrameError(IsoTpError):
    """
    Happens when the reception of a multi packet message reception is interrupted with a new First Frame PDU.
    """
    pass
class WrongSequenceNumberError(IsoTpError):
    """
    Happens when a consecutive frame is received with a wrong sequence number.
    """
    pass
class UnsuportedWaitFrameError(IsoTpError):
    """
    Happens when a Flow Control PDU with FlowStatus=Wait is received and :ref:`wftmax<param_wftmax>` is set to 0
    """
    pass
class MaximumWaitFrameReachedError(IsoTpError):
    """
    Happens when too much Flow Control PDU with FlowStatus=Wait is received. Refer to :ref:`wftmax<param_wftmax>`
    """
    pass


from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
  name = 'can-isotp',
  packages = find_packages(),
  package_data={
    '': ['*.conf'],
  },
  version = '1.1',
  description = 'Module enabling the IsoTP protocol defined by ISO-15765',
  long_description=long_description,
  author = 'Pier-Yves Lessard',
  author_email = 'py.lessard@gmail.com',
  license='MIT',
  url = 'https://github.com/pylessard/python-can-isotp',
  download_url = 'https://github.com/pylessard/python-can-isotp/archive/v1.1.tar.gz',
  keywords = ['isotp', 'can', 'iso-15765', '15765', 'iso15765'], 
  python_requires='>=3',
  classifiers = [
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
        "Operating System :: POSIX :: Linux",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
        ],
)


class AddressingMode:
    Normal_11bits = 0
    Normal_29bits = 1
    NormalFixed_29bits = 2
    Extended_11bits = 3
    Extended_29bits = 4
    Mixed_11bits = 5
    Mixed_29bits = 6

    @classmethod
    def get_name(cls, num):
        if num == cls.Normal_11bits:
            return 'Normal_11bits'
        if num == cls.Normal_29bits:
            return 'Normal_29bits'
        if num == cls.NormalFixed_29bits:
            return 'NormalFixed_29bits'
        if num == cls.Extended_11bits:
            return 'Extended_11bits'
        if num == cls.Extended_29bits:
            return 'Extended_29bits'
        if num == cls.Mixed_11bits:
            return 'Mixed_11bits'
        if num == cls.Mixed_29bits:
            return 'Mixed_29bits'

        return 'Unknown'

class TargetAddressType:
    Physical = 0        # 1 to 1 communication
    Functional = 1      # 1 to n communication

class Address:
    """
    Represents the addressing information (N_AI) of the IsoTP layer. Will define what messages will be received and how to craft transmitted message to reach a specific party.

    Parameters must be given according to the addressing mode. When not needed, a parameter may be left unset or set to ``None``.

    Both the :class:`TransportLayer<isotp.TransportLayer>` and the :class:`isotp.socket<isotp.socket>` expects this address object

    :param addressing_mode: The addressing mode. Valid values are defined by the :class:`AddressingMode<isotp.AddressingMode>` class
    :type addressing_mode: int

    :param txid: The CAN ID for transmission. Used for these addressing mode: ``Normal_11bits``, ``Normal_29bits``, ``Extended_11bits``, ``Extended_29bits``, ``Mixed_11bits``
    :type txid: int or None

    :param rxid: The CAN ID for reception. Used for these addressing mode: ``Normal_11bits``, ``Normal_29bits``, ``Extended_11bits``, ``Extended_29bits``, ``Mixed_11bits``
    :type rxid: int or None

    :param target_address: Target address (N_TA) used in ``NormalFixed_29bits`` and ``Mixed_29bits`` addressing mode.
    :type target_address: int or None

    :param source_address: Source address (N_SA) used in ``NormalFixed_29bits`` and ``Mixed_29bits`` addressing mode.
    :type source_address: int or None

    :param address_extension: Address extension (N_AE) used in ``Mixed_11bits``, ``Mixed_29bits`` addressing mode
    :type address_extension: int or None
    """

    def __init__(self, addressing_mode = AddressingMode.Normal_11bits, txid=None, rxid=None, target_address=None, source_address=None, address_extension=None):

        self.addressing_mode    = addressing_mode
        self.target_address     = target_address
        self.source_address     = source_address
        self.address_extension  = address_extension
        self.txid               = txid
        self.rxid               = rxid
        self.is_29bits          = True if self.addressing_mode in [ AddressingMode.Normal_29bits, AddressingMode.NormalFixed_29bits, AddressingMode.Extended_29bits, AddressingMode.Mixed_29bits] else False

        self.validate()

        # From here, input is good. Do some precomputing for speed optimization without bothering about types or values
        self.tx_arbitration_id_physical     = self._get_tx_arbitraton_id(TargetAddressType.Physical)
        self.tx_arbitration_id_functional   = self._get_tx_arbitraton_id(TargetAddressType.Functional)
       
        self.rx_arbitration_id_physical     = self._get_rx_arbitration_id(TargetAddressType.Physical)
        self.rx_arbitration_id_functional   = self._get_rx_arbitration_id(TargetAddressType.Functional)

        self.tx_payload_prefix = bytearray()
        self.rx_prefix_size = 0

        if self.addressing_mode in [AddressingMode.Extended_11bits, AddressingMode.Extended_29bits]:
            self.tx_payload_prefix.extend(bytearray([self.target_address]))
            self.rx_prefix_size = 1
        elif self.addressing_mode in [AddressingMode.Mixed_11bits, AddressingMode.Mixed_29bits]:
            self.tx_payload_prefix.extend(bytearray([self.address_extension]))
            self.rx_prefix_size = 1

        self.rxmask = None
        if self.addressing_mode == AddressingMode.NormalFixed_29bits:
            self.rxmask = 0x18DA0000    # This should ignore variant between Physical and Functional addressing
        elif self.addressing_mode == AddressingMode.Mixed_29bits:
            self.rxmask = 0x18CD0000    # This should ignore variant between Physical and Functional addressing

        if self.addressing_mode in [AddressingMode.Normal_11bits, AddressingMode.Normal_29bits]:
            self.is_for_me = self._is_for_me_normal
        elif self.addressing_mode in [AddressingMode.Extended_11bits, AddressingMode.Extended_29bits]:
            self.is_for_me = self._is_for_me_extended
        elif self.addressing_mode == AddressingMode.NormalFixed_29bits:
            self.is_for_me = self._is_for_me_normalfixed
        elif self.addressing_mode == AddressingMode.Mixed_11bits:
            self.is_for_me = self._is_for_me_mixed_11bits
        elif self.addressing_mode == AddressingMode.Mixed_29bits:
            self.is_for_me = self._is_for_me_mixed_29bits
        else:
            raise RuntimeError('This exception should never be raised.')

    def validate(self):
        if self.addressing_mode not in [AddressingMode.Normal_11bits,AddressingMode.Normal_29bits,AddressingMode.NormalFixed_29bits,AddressingMode.Extended_11bits,AddressingMode.Extended_29bits,AddressingMode.Mixed_11bits,AddressingMode.Mixed_29bits]:
            raise ValueError('Addressing mode is not valid')

        if self.addressing_mode in [AddressingMode.Normal_11bits, AddressingMode.Normal_29bits]:
            if self.rxid is None or self.txid is None:
                raise ValueError('txid and rxid must be specified for Normal addressing mode (11 bits ID)')
            if self.rxid == self.txid:
                raise ValueError('txid and rxid must be different for Normal addressing mode')
        
        elif self.addressing_mode == AddressingMode.NormalFixed_29bits:
            if self.target_address is None or self.source_address is None:
                raise ValueError('target_address and source_address must be specified for Normal Fixed addressing (29 bits ID)')
        
        elif self.addressing_mode in [AddressingMode.Extended_11bits, AddressingMode.Extended_29bits]:
            if self.target_address is None or self.rxid is None or self.txid is None:
                raise ValueError('target_address, rxid and txid must be specified for Extended addressing mode ')
            if self.rxid == self.txid:
                raise ValueError('txid and rxid must be different')

        elif self.addressing_mode == AddressingMode.Mixed_11bits:
            if self.rxid is None or self.txid is None or self.address_extension is None:
                raise ValueError('rxid, txid and address_extension must be specified for Mixed addressing mode')

        elif self.addressing_mode == AddressingMode.Mixed_29bits:
            if self.target_address is None or self.source_address is None or self.address_extension is None:
                raise ValueError('target_address, source_address and address_extension must be specified for Mixed addressing mode')

        if self.target_address is not None:
            if not isinstance(self.target_address, int):
                raise ValueError('target_address must be an integer')
            if self.target_address < 0 or self.target_address > 0xFF:
                raise ValueError('target_address must be an integer between 0x00 and 0xFF')

        if self.source_address is not None:
            if not isinstance(self.source_address, int):
                raise ValueError('source_address must be an integer')
            if self.source_address < 0 or self.source_address > 0xFF:
                raise ValueError('source_address must be an integer between 0x00 and 0xFF')

        if self.address_extension is not None:
            if not isinstance(self.address_extension, int):
                raise ValueError('source_address must be an integer')
            if self.address_extension < 0 or self.address_extension > 0xFF:
                raise ValueError('address_extension must be an integer between 0x00 and 0xFF')

        if self.txid is not None:
            if not isinstance(self.txid, int):
                raise ValueError('txid must be an integer')
            if self.txid < 0:
                raise ValueError('txid must be greater than 0')
            if not self.is_29bits:
                if self.txid > 0x7FF:
                    raise ValueError('txid must be smaller than 0x7FF for 11 bits identifier')

        if self.rxid is not None:
            if not isinstance(self.rxid, int):
                raise ValueError('rxid must be an integer')
            if self.rxid < 0:
                raise ValueError('rxid must be greater than 0')
            if not self.is_29bits:
                if self.rxid > 0x7FF:
                    raise ValueError('rxid must be smaller than 0x7FF for 11 bits identifier')

    def get_tx_arbitraton_id(self, address_type=TargetAddressType.Physical):
        if address_type == TargetAddressType.Physical:
            return self.tx_arbitration_id_physical
        else:
            return self.tx_arbitration_id_functional

    def get_rx_arbitraton_id(self, address_type=TargetAddressType.Physical):
        if address_type == TargetAddressType.Physical:
            return self.rx_arbitration_id_physical
        else:
            return self.rx_arbitration_id_functional

    def _get_tx_arbitraton_id(self, address_type):
        if self.addressing_mode == AddressingMode.Normal_11bits:
            return self.txid
        elif self.addressing_mode == AddressingMode.Normal_29bits:
            return self.txid
        elif self.addressing_mode == AddressingMode.NormalFixed_29bits:
            bits23_16 = 0xDA0000 if address_type==TargetAddressType.Physical else 0xDB0000
            return 0x18000000 | bits23_16 | (self.target_address << 8) | self.source_address
        elif self.addressing_mode == AddressingMode.Extended_11bits:
            return self.txid
        elif self.addressing_mode == AddressingMode.Extended_29bits:
            return self.txid
        elif self.addressing_mode == AddressingMode.Mixed_11bits:
            return self.txid
        elif self.addressing_mode == AddressingMode.Mixed_29bits:
            bits23_16 = 0xCE0000 if address_type==TargetAddressType.Physical else 0xCD0000
            return 0x18000000 | bits23_16 | (self.target_address << 8) | self.source_address

    def _get_rx_arbitration_id(self, address_type=TargetAddressType.Physical):
        if self.addressing_mode == AddressingMode.Normal_11bits:
            return self.rxid
        elif self.addressing_mode == AddressingMode.Normal_29bits:
            return self.rxid
        elif self.addressing_mode == AddressingMode.NormalFixed_29bits:
            bits23_16 = 0xDA0000 if address_type==TargetAddressType.Physical else 0xDB0000
            return 0x18000000 | bits23_16 | (self.source_address << 8) | self.target_address
        elif self.addressing_mode == AddressingMode.Extended_11bits:
            return self.rxid
        elif self.addressing_mode == AddressingMode.Extended_29bits:
            return self.rxid
        elif self.addressing_mode == AddressingMode.Mixed_11bits:
            return self.rxid
        elif self.addressing_mode == AddressingMode.Mixed_29bits:
            bits23_16 = 0xCE0000 if address_type==TargetAddressType.Physical else 0xCD0000
            return 0x18000000 | bits23_16 | (self.source_address << 8) | self.target_address

    def _is_for_me_normal(self, msg):
        if self.is_29bits == msg.is_extended_id:
            return msg.arbitration_id == self.rxid
        return False

    def _is_for_me_extended(self, msg):
        if self.is_29bits == msg.is_extended_id:
            if msg.data is not None and len(msg.data) > 0:
                return msg.arbitration_id == self.rxid and int(msg.data[0]) == self.source_address 
        return False

    def _is_for_me_normalfixed(self, msg):
        if self.is_29bits == msg.is_extended_id:
            return ((msg.arbitration_id >> 16) & 0xFF) in [218,219] and (msg.arbitration_id & 0xFF00) >> 8 == self.source_address and msg.arbitration_id & 0xFF == self.target_address
        return False

    def _is_for_me_mixed_11bits(self, msg):
        if self.is_29bits == msg.is_extended_id:
            if msg.data is not None and len(msg.data) > 0:
                return msg.arbitration_id == self.rxid and int(msg.data[0]) == self.address_extension
        return False

    def _is_for_me_mixed_29bits(self, msg):
        if self.is_29bits == msg.is_extended_id:
            if msg.data is not None and len(msg.data) > 0:
                return ((msg.arbitration_id >> 16) & 0xFF) in [205,206] and (msg.arbitration_id & 0xFF00) >> 8 == self.source_address and msg.arbitration_id & 0xFF == self.target_address and int(msg.data[0]) == self.address_extension
        return False

    def requires_extension_byte(self):
        return True if self.addressing_mode in [AddressingMode.Extended_11bits, AddressingMode.Extended_29bits, AddressingMode.Mixed_11bits, AddressingMode.Mixed_29bits] else False

    def get_tx_extension_byte(self):
        if self.addressing_mode in [AddressingMode.Extended_11bits, AddressingMode.Extended_29bits]:
            return self.target_address
        if self.addressing_mode in [ AddressingMode.Mixed_11bits, AddressingMode.Mixed_29bits]:
            return self.address_extension

    def get_rx_extension_byte(self):
        if self.addressing_mode in [AddressingMode.Extended_11bits, AddressingMode.Extended_29bits]:
            return self.source_address
        if self.addressing_mode in [AddressingMode.Mixed_11bits, AddressingMode.Mixed_29bits]:
            return self.address_extension

    def get_content_str(self):
        val_dict = {}
        keys =  ['target_address', 'source_address', 'address_extension', 'txid', 'rxid']
        for key in keys:
            val = getattr(self, key)
            if val is not None:
                val_dict[key] = val
        vals_str = ', '.join(['%s:0x%02x' % (k,val_dict[k]) for k in val_dict])
        return '[%s - %s]' % (AddressingMode.get_name(self.addressing_mode), vals_str)
    
    def __repr__(self):
        return '<IsoTP Address %s at 0x%08x>' % (self.get_content_str(), id(self))



import unittest
import queue



# Just a class with some helper such as simulate_rx() to make the tests cleaners.
class TransportLayerBaseTest(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        unittest.TestCase.__init__(self, *args, **kwargs)
        self.ll_rx_queue = queue.Queue()
        self.ll_tx_queue = queue.Queue()
        self.error_triggered = {}

    def error_handler(self, error):
        if error.__class__ not in self.error_triggered:
            self.error_triggered[error.__class__] = []

        self.error_triggered[error.__class__].append(error)

    def stack_txfn(self, msg):
        if not self.ll_tx_queue.full():
            self.ll_tx_queue.put(msg)

    def stack_rxfn(self):
        if not self.ll_rx_queue.empty():
            return  self.ll_rx_queue.get()

    def rx_isotp_frame(self):
        return self.stack.recv()

    def tx_isotp_frame(self, frame):
        self.stack.send(frame)

    def get_tx_can_msg(self):
        if not self.ll_tx_queue.empty():
            return self.ll_tx_queue.get()

    def make_payload(self, size, start_val=0):
        return [int(x%0x100) for x in range(start_val, start_val+size)]

    def assert_sent_flow_control(self, stmin, blocksize, prefix = None, padding_byte=None, extra_msg=''):
        msg = self.get_tx_can_msg()
        self.assertIsNotNone(msg, 'Expected a Flow Control message, but none was sent.' + ' ' + extra_msg)
        data = bytearray()
        if prefix is not None:
            data.extend(prefix)

        data.extend(bytearray([0x30, blocksize, stmin]))
        if padding_byte is not None:
            padlen = 5
            if prefix is not None:
                padlen -= len(prefix)
            data.extend(bytearray([padding_byte]*padlen))

        self.assertEqual(msg.data, data, 'Message sent is not the wanted flow control'+ ' ' + extra_msg)    # Flow Control
        self.assertEqual(msg.dlc, len(data), 'Flow control message has wrong DLC. Expecting=0x%02x, received=0x%02x' % (len(data), msg.dlc))    

    def assert_error_triggered(self, error_type):
        if error_type in self.error_triggered:
            if len(self.error_triggered[error_type]):
                return

        raise AssertionError('Error of type %s not triggered' % error_type.__name__)

    def assert_no_error_triggered(self):
        if len(self.error_triggered) > 0:
            raise AssertionError('%d errors hsa been triggered while non should have' % len(self.error_triggered))          

    def clear_errors(self):
        self.error_triggered = {}

    def init_test_case(self):
        while not self.ll_rx_queue.empty():
            self.ll_rx_queue.get()

        while not self.ll_tx_queue.empty():
            self.ll_tx_queue.get()

        self.clear_errors()

    def simulate_rx_msg(self, msg):
        self.ll_rx_queue.put(msg)

    def make_flow_control_data(self, flow_status, stmin, blocksize, prefix=None):
        data = bytearray()
        if prefix is not None:
            data.extend(bytearray(prefix))
        data.extend(bytearray( [0x30 | (flow_status & 0xF), blocksize, stmin]))

        return data